import time
import torch
import random
from tqdm import tqdm
import utils.utilis_func as uf
import numpy as np
import qm9.utils as qm9utils
import qm9.visualizer as vis
from utils.sampling import sample
from qm9.sampling import sample_chain, sample_sweep_conditional
from qm9.analyze import analyze_stability_for_molecules

# from traintest.sampling import sample_chain, sample, sample_sweep_conditional

def prepare_batch_data(args, data, device, dtype, partition='Train'):
    x = data['positions'].to(device, dtype)
    node_mask = data['atom_mask'].to(device, dtype).unsqueeze(2)
    edge_mask = data['edge_mask'].to(device, dtype)
    one_hot = data['one_hot'].to(device, dtype)
    charges = (data['charges'] if args.include_charges else torch.zeros(0)).to(device, dtype)
    try:
        noise = data['noise'].to(device, dtype)
    except:
        noise = None
    x = uf.remove_mean_with_mask(x, node_mask)

    if args.augment_noise > 0:
        # Add noise eps ~ N(0, augment_noise) around points.
        eps = uf.sample_center_gravity_zero_gaussian_with_mask(x.size(), x.device, node_mask)
        x = x + eps * args.augment_noise

    x = uf.remove_mean_with_mask(x, node_mask)
    if args.data_augmentation:
        x = uf.random_rotation(x).detach()

    # check_mask_correct([x, one_hot, charges], node_mask)
    uf.assert_mean_zero_with_mask(x, node_mask)

    h = {'categorical': one_hot, 'integer': charges}
    return x, h, node_mask, edge_mask, noise


def train_epoch(args, loader, epoch, model, model_ema, ema, device, dtype, property_norms, optim,
                nodes_dist, gradnorm_queue, dataset_info, prop_dist, rank):
    model.train()
    nll_epoch = []
    n_iterations = len(loader)
    if rank == 0:
        loader = tqdm(loader, ncols=80)
    loss = 0
    for i, data in enumerate(loader):
        x, h, node_mask, edge_mask, noise = prepare_batch_data(args, data, device, dtype)
        if len(args.conditioning) > 0:
            context = qm9utils.prepare_context(args.conditioning, data, property_norms).to(device, dtype)
            assert_correctly_masked(context, node_mask)
        else:
            context = None
        optim.zero_grad()
        nll, reg_term, mean_abs_z = compute_loss_and_nll(args, model, nodes_dist, x, h, node_mask, edge_mask, context, noise)
        loss = nll
        loss.backward()
        if args.clip_grad:
            grad_norm = uf.gradient_clipping(model, gradnorm_queue)
        else:
            grad_norm = 0.
        optim.step()
        # Update EMA if enabled.
        if args.ema_decay > 0:
            ema.update_model_average(model_ema, model)
        if rank == 0:
            train_des = (f"\rE: {epoch}, i: {i}/{n_iterations}, "
                         f"Loss {loss.item():.2f}, Size: {x.size(1)}, "
                         f"GN: {grad_norm:.2f}")
            loader.set_description(train_des)
            nll_epoch.append(nll.item())
        if args.break_train_epoch:
            break
    return loss.item()

def compute_transport_loss(args, loader, epoch, model, model_ema, ema, device, dtype, property_norms, optim,
                nodes_dist, gradnorm_queue, dataset_info, prop_dist, rank):
    model.eval()
    if rank == 0:
        loader = tqdm(loader, ncols=80)
    total_cost = 0
    total_atoms = 0
    for i, data in enumerate(loader):
        x, h, node_mask, edge_mask, noise = prepare_batch_data(args, data, device, dtype)
        if len(args.conditioning) > 0:
            context = qm9utils.prepare_context(args.conditioning, data, property_norms).to(device, dtype)
            assert_correctly_masked(context, node_mask)
        else:
            context = None
        transport_cost, atoms, mean_abs_z = compute_loss_and_nll(args, model, nodes_dist, x, h, node_mask, edge_mask, context, noise)
        total_cost += transport_cost
        total_atoms += atoms
    return total_cost.item(), total_atoms.item()


def check_mask_correct(variables, node_mask):
    for i, variable in enumerate(variables):
        if len(variable) > 0:
            assert_correctly_masked(variable, node_mask)


def test(args, loader, epoch, eval_model, device, dtype, property_norms, nodes_dist, rank, partition='Test'):
    eval_model.eval()
    with torch.no_grad():
        nll_epoch = 0
        n_samples = 0

        n_iterations = len(loader)
        if rank == 0:
            loader = tqdm(loader, ncols=80)
        for i, data in enumerate(loader):
            x, h, node_mask, edge_mask, noise = prepare_batch_data(args, data, device, dtype, partition)
            batch_size = x.size(0)
            if len(args.conditioning) > 0:
                context = qm9utils.prepare_context(args.conditioning, data, property_norms).to(device, dtype)
                uf.assert_correctly_masked(context, node_mask)
            else:
                context = None

            # transform batch through flow
            nll, reg_term, mean_abs_z = compute_loss_and_nll(args, eval_model, nodes_dist, x, h, node_mask, edge_mask, context, noise)
            # standard nll from forward KL

            nll_epoch += nll.item() * batch_size
            n_samples += batch_size
            if rank == 0:
                train_des = (f"\r {partition} \t e: {epoch}, i: {i}/{n_iterations}, "
                             f"NLL: {nll_epoch/n_samples:.2f}")
                loader.set_description(train_des)

    return nll_epoch/n_samples


def save_and_sample_chain(model, args, device, dataset_info, prop_dist,
                          epoch=0, id_from=0, batch_id=''):
    one_hot, charges, x = sample_chain(args=args, device=device, flow=model,
                                       n_tries=1, dataset_info=dataset_info, prop_dist=prop_dist)

    vis.save_xyz_file(f'outputs/{args.exp_name}/epoch_{epoch}_{batch_id}/chain/',
                      one_hot, charges, x, dataset_info, id_from, name='chain')

    return one_hot, charges, x


def sample_different_sizes_and_save(model, nodes_dist, args, device, dataset_info, prop_dist,
                                    n_samples=5, epoch=0, batch_size=100, batch_id=''):
    batch_size = min(batch_size, n_samples)
    for counter in range(int(n_samples/batch_size)):
        nodesxsample = nodes_dist.sample(batch_size)
        one_hot, charges, x, node_mask = sample(args, device, model, prop_dist=prop_dist,
                                                nodesxsample=nodesxsample,
                                                dataset_info=dataset_info)
        print(f"Generated molecule: Positions {x[:-1, :, :]}")
        vis.save_xyz_file(f'outputs/{args.exp_name}/epoch_{epoch}_{batch_id}/', one_hot, charges, x, dataset_info,
                          batch_size * counter, name='molecule')


def analyze_and_save(epoch, model_sample, nodes_dist, args, device, dataset_info, prop_dist, property_norms,
                     n_samples=1000, batch_size=100):
    print(f'Analyzing molecule stability at epoch {epoch}...')
    batch_size = min(batch_size, n_samples)
    assert n_samples % batch_size == 0
    molecules = {'one_hot': [], 'x': [], 'node_mask': []}
    loss = 0
    for i in range(int(n_samples/batch_size)):
        nodesxsample = nodes_dist.sample(batch_size)
        one_hot, charges, x, node_mask, loss = sample(args, device, model_sample, dataset_info, prop_dist,
                                                nodesxsample=nodesxsample, property_norms=property_norms)

        molecules['one_hot'].append(one_hot.detach().cpu())
        molecules['x'].append(x.detach().cpu())
        molecules['node_mask'].append(node_mask.detach().cpu())

    molecules = {key: torch.cat(molecules[key], dim=0) for key in molecules}
    validity_dict, rdkit_tuple = analyze_stability_for_molecules(molecules, dataset_info)
    rdkit_info = rdkit_tuple[0]
    metrics = torch.FloatTensor([validity_dict['atm_stable'], validity_dict['mol_stable'], rdkit_info[0], rdkit_info[0] * rdkit_info[1], rdkit_info[2], loss]).to(device)
    return metrics


def save_and_sample_conditional(args, device, model, data, prop_dist, dataset_info, epoch=0, id_from=0):
    one_hot, charges, x, node_mask = sample_sweep_conditional(args, device, model, data, dataset_info, prop_dist)

    vis.save_xyz_file(
        'outputs/%s/epoch_%d/conditional/' % (args.exp_name, epoch), one_hot, charges, x, dataset_info,
        id_from, name='conditional', node_mask=node_mask)

    return one_hot, charges, x


def sum_except_batch(x):
    return x.view(x.size(0), -1).sum(dim=-1)


def assert_correctly_masked(variable, node_mask):
    assert (variable * (1 - node_mask)).abs().sum().item() < 1e-8


def compute_loss_and_nll(args, generative_model, nodes_dist, x, h, node_mask, edge_mask, context, noise):
    bs, n_nodes, n_dims = x.size()
    if args.probabilistic_model == 'diffusion':
        edge_mask = edge_mask.view(bs, n_nodes * n_nodes)

        assert_correctly_masked(x, node_mask)

        # Here x is a position tensor, and h is a dictionary with keys
        # 'categorical' and 'integer'.
        nll = generative_model(x, h, node_mask, edge_mask, context, noise)

        # Average over batch.
        nll = nll.mean(0)

        reg_term = torch.tensor([0.0]).to(nll.device)
        mean_abs_z = 0.0
    elif args.probabilistic_model == 'vae':
        edge_mask = edge_mask.view(bs, n_nodes * n_nodes)
        nll = generative_model(x, h, node_mask, edge_mask, context)
        nll = nll.mean(0)
        reg_term = torch.tensor([0.0]).to(nll.device)
        mean_abs_z = 0.0
    elif args.probabilistic_model == 'transport':
        edge_mask = edge_mask.view(bs, n_nodes * n_nodes)

        assert_correctly_masked(x, node_mask)

        # Here x is a position tensor, and h is a dictionary with keys
        # 'categorical' and 'integer'.
        nll = generative_model.compute_transport_cost(x, h, node_mask, edge_mask, context, noise)

        # Average over batch.
        nll = nll.sum(0)

        reg_term = node_mask.sum()
        mean_abs_z = 0.0
    else:
        raise ValueError(args.probabilistic_model)

    return nll, reg_term, mean_abs_z